/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 2001-2008 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

/* match.c implements pattern matching (globbing) */

#include <config.h>

#include <sys/types.h>	/* clex.h */
#include <string.h>		/* strlen() */

#include "clex.h"
#include "match.h"

#include "lex.h"		/* cmd2lex() */
#include "sdstring.h"	/* sd_copy() */
#include "util.h"		/* str_compare() */

static const char *expr;	/* shell regular expression */
static char *sym;			/* corresponding string of LEX_XXX */
static size_t expr_len;		/* length of regular expression */
static FLAG allfiles;		/* trivial case - match everything */
static FLAG match_dot;		/* literal match for leading dot */

/* match result - used in check_match() only */
#define MATCH_NO				1
#define MATCH_YES				2
#define MATCH_NEVER				3

static int
rangecmp(int from, int ch, int to)
{
	static char f[2] = "?", c[2] = "?", t[2] = "?";

	f[0] = (char)from;
	c[0] = (char)ch;
	t[0] = (char)to;
	return str_compare(f,c) <= 0 && str_compare(c,t) <= 0;
}

/*
 * check if shell regular expression set by last successful match_set_pattern()
 * call and starting at position 'i' matches the string 'word'
 */
static int
check_match(size_t i, const char *word)
{
	int m, from;
	FLAG inverse, listmatch;
	char ch;

	from = 0; ch = '\0';		/* to prevent compiler warning */
	inverse = listmatch = 0;	/* to prevent compiler warning */

	for (; i < expr_len; i++)
		switch (sym[i]) {
		case LEX_QMARK:
		case LEX_MARK:
			/* ignore */
			break;
		case LEX_MATCH_ONE:
			if (*word++ == '\0')
				return MATCH_NO;
			break;
		case LEX_MATCH_ANY:
			if (i + 1 == expr_len)
				/* shortcut: single * matches it all */
				return MATCH_YES;
			do {
				if ( (m = check_match(i + 1,word)) != MATCH_NO)
					return m;
			} while (*word++);
			/*
			 * match impossible, stop recursion, otherwise you can
			 * spend many hours looping here if there is a lot of asterisks
			 */
			return MATCH_NEVER;
		case LEX_MATCH_BEGIN:
			inverse = listmatch = 0;
			ch = *word++;	/* char to match */
			break;
		case LEX_MATCH_INVERSE:
			inverse = 1;
			break;
		case LEX_MATCH_LIST:
			if (expr[i] == ch)
				listmatch = 1;
			break;
		case LEX_MATCH_RANGE_FROM:
			from = expr[i];
			break;
		case LEX_MATCH_RANGE_TO:
			if (rangecmp(from,ch,expr[i]))
				listmatch = 1;
			break;
		case LEX_MATCH_END:
			if (listmatch == inverse)
				return MATCH_NO;
			break;
		default:
			/* literal */
			if (expr[i] != *word++)
				return MATCH_NO;
		}

	return *word == '\0' ? MATCH_YES : MATCH_NO ;
}

int
match(const char *word)
{
	return allfiles || (*word == '.' && !match_dot && !FOPT(FOPT_ALL)
	  ? 0 : check_match(0,word) == MATCH_YES);
}

int
match_set_pattern(const char *pattern, const char *lex)
{
	static SDSTRING expression = { 0,"" }, symbols = { 0,"" };
	int i;

	if ( (allfiles = strcmp(pattern,"*") == 0 && FOPT(FOPT_ALL)) )
		return 0;

	sd_copy(&expression,pattern);
	expr = SDSTR(expression);
	expr_len = strlen(expr);

	if (lex == 0)
		lex = cmd2lex(pattern);
	if (lex[expr_len] != LEX_END_OK) {
		lex_error("PATTERN CHECK",lex[expr_len]);
		return -1;
	}
	sd_copy(&symbols,lex);
	sym = SDSTR(symbols);

	for (i = 0; sym[i] == LEX_QMARK; i++)
		;
	match_dot = expr[i] == '.';

	return 0;
}
