/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 2001-2008 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <config.h>

#include <sys/types.h>	/* clex.h */
#include <sys/stat.h>	/* stat() */
#include <fcntl.h>		/* open() */
#include <stdlib.h>		/* qsort() */
#include <string.h>		/* strcpy() */
#include <unistd.h>		/* close() */

#include "clex.h"
#include "select.h"

#include "control.h"	/* get_current_mode() */
#include "edit.h"		/* edit_putstr() */
#include "inout.h"		/* win_panel() */
#include "match.h"		/* match() */
#include "list.h"		/* list_both_directories() */
#include "sdstring.h"	/* SDSTR() */
#include "signals.h"	/* signal_ctrlc_on() */
#include "sort.h"		/* sort_files() */
#include "ustring.h"	/* USTR() */
#include "util.h"		/* pathname_join() */

#define FN_SELECT	1
#define FN_DESELECT	2
#define FN_TOGGLE	3

static SDSTRING sel_pat = { 0,"*" }, desel_pat = { 0,"*" };
static FLAG mode_sel;		/* 1 if MODE_SELECT, 0 if MODE_DESELECT */

void
select_prepare(void)
{
	static FLAG hint = 1;
	const char *prompt, *pat;
	size_t len;

	mode_sel = get_current_mode() == MODE_SELECT;

	panel = ppanel_file->pd;
	if (panel->filtering == 1)
		panel->filtering = 2;

	prompt = "DESELECT files: " + (mode_sel ? 2 : 0);
	edit_setprompt(&line_tmp,prompt);
	textline = &line_tmp;
	edit_nu_putstr(pat = mode_sel ? SDSTR(sel_pat) : SDSTR(desel_pat));
	len = strlen(pat);
	if (len >= 2 && pat[len - 2] == '/' && pat[len - 1] == 'a')
		textline->curs -= 2;

	if (TCLR(hint))
		win_remark("wildcards: ? * and [..], see help");
}

void
compare_prepare(void)
{
	/* leave cursor position unchanged */
	panel = panel_compare.pd;
	textline = 0;

	if (strcmp(USTR(ppanel_file->dir),USTR(ppanel_file->other->dir)) == 0)
		win_remark("COMPARE: the directory is the same in both panels");
}

void
compare_summary_prepare(void)
{
	panel_cmp_sum.pd->top = panel_cmp_sum.pd->curs = panel_cmp_sum.pd->min;
	panel_cmp_sum.pd->cnt = panel_cmp_sum.errors ? 6 : 5;
	panel = panel_cmp_sum.pd;
	textline = 0;
}

static int
selectfile(FILE_ENTRY *pfe, int function)
{
	if (!pfe->select) {
		if (function != FN_DESELECT) {
			pfe->select = 1;	/* one selected entry more */
			return 1;
		}
	}
	else if (function != FN_SELECT) {
		pfe->select = 0;
		return -1;				/* one less */
	}
	return 0;					/* no change */
}

void
cx_select_toggle(void)
{
	ppanel_file->selected +=
	  selectfile(ppanel_file->files[ppanel_file->pd->curs],FN_TOGGLE);

	/* cursor down */
	if (ppanel_file->pd->curs < ppanel_file->pd->cnt - 1) {
		ppanel_file->pd->curs++;
		LIMIT_MIN(ppanel_file->pd->top,
		  ppanel_file->pd->curs - display.panlines + 1);
	}

	win_panel_opt();
}

void
cx_select_range(void)
{
	int i, s, fn;

	i = ppanel_file->pd->curs;
	fn = ppanel_file->files[i]->select ? FN_DESELECT : FN_SELECT;
	while (i >= 0) {
		if ( (s = selectfile(ppanel_file->files[i--],fn)) == 0)
			break;
		ppanel_file->selected += s;
	}
	win_panel();
}

void cx_select_invert(void)
{
	int i;

	for (i = 0; i < ppanel_file->pd->cnt; i++)
		TOGGLE(ppanel_file->files[i]->select);
	ppanel_file->selected = ppanel_file->pd->cnt - ppanel_file->selected;
	win_panel();
}

void
cx_select_files(void)
{
	int i, cnt;
	const char *sre;
	FILE_ENTRY *pfe;

	if (line_tmp.size == 0) {
		next_mode = MODE_SPECIAL_RETURN;
		return;
	}
	sre = USTR(textline->line);
	if (match_set_pattern(sre,0) < 0)
		return;	/* user should fix the pattern */

	/* save the pattern */
	sd_copy(mode_sel ? &sel_pat : &desel_pat,sre);

	for (i = cnt = 0; i < ppanel_file->pd->cnt; i++) {
		pfe = ppanel_file->files[i];
		if (pfe->select != mode_sel && match(SDSTR(pfe->file))) {
			pfe->select = mode_sel;
			cnt++;
		}
	}
	ppanel_file->selected += mode_sel ? +cnt : -cnt;
	win_panel();
	next_mode = MODE_SPECIAL_RETURN;
}

static int
qcmp(const void *e1, const void *e2)
{
	return strcmp( /* not strcoll() */
	  SDSTR((*(FILE_ENTRY **)e1)->file),
	  SDSTR((*(FILE_ENTRY **)e2)->file));
}

#define CMP_BUF_STR	16384

/* return value: -1 error, 0 compare ok, +1 compare failed */
static int
data_cmp(int fd1, int fd2)
{
	struct stat st1, st2;
	static char buff1[CMP_BUF_STR], buff2[CMP_BUF_STR];
	off_t filesize;
	size_t chunksize;

	if (fstat(fd1,&st1) < 0 || !S_ISREG(st1.st_mode))
		return -1;
	if (fstat(fd2,&st2) < 0 || !S_ISREG(st2.st_mode))
		return -1;
	if (st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino)
		/* same file */
		return 0;
	if ((filesize = st1.st_size) != st2.st_size)
		return 1;

	while (filesize > 0) {
		chunksize = filesize > CMP_BUF_STR ? CMP_BUF_STR : filesize;
		if (ctrlc_flag || read_fd(fd1,buff1,chunksize) != chunksize
		  || read_fd(fd2,buff2,chunksize) != chunksize)
			return -1;
		if (memcmp(buff1,buff2,chunksize) != 0)
			return 1;
		filesize -= chunksize;
	}

	return 0;
}

/* return value: -1 error, 0 compare ok, +1 compare failed */
static int
file_cmp(const char *file1, const char *file2)
{
	int cmp, fd1, fd2;

	fd1 = open(file1,O_RDONLY | O_NONBLOCK);
	if (fd1 < 0)
		return -1;
	fd2 = open(file2,O_RDONLY | O_NONBLOCK);
	if (fd2 < 0) {
		close(fd1);
		return -1;
	}
	cmp = data_cmp(fd1,fd2);
	close(fd1);
	close(fd2);
	return cmp;
}

/*
 * compare panels
 * levels:
 *	0: name, type       (symbolic links ok)
 *	1: name, type, size (symbolic links ok)
 *	2: name, type, size (exactly)
 *	3: name, type, size, ownership&mode
 *	4: name, type, size, contents
 *	5: name, type, size, ownership&mode, contents
 */
static void
compare_panels(int level)
{
	int min, med, max, cmp, i, cnt1;
	const char *name2;
	FILE_ENTRY *pfe1, *pfe2;
	static FILE_ENTRY **p1 = 0;	/* copy of panel1, sorted for binary search */
	static int p1_alloc = 0;

	/*
	 * - select all files and both panels
	 * - for each file from panel #1 (pfe1):
	 *		- find a matching file from #2 (pfe2)
	 *		- if a pair is found, compare them according to the selected level
	 *		- if the comapred files are equal, deselect them
	 */

	panel_cmp_sum.errors = panel_cmp_sum.names = panel_cmp_sum.equal = 0;

	/* reread panels */
	list_both_directories();

	ctrlc_flag = 0;
	if (level >= 4) {	/* data will be compared */
		signal_ctrlc_on();
		win_waitmsg();
		pathname_set_directory(USTR(ppanel_file->other->dir));
	}

	if ( (cnt1 = ppanel_file->pd->cnt) ) {
		if (p1_alloc < cnt1) {
			p1_alloc = cnt1;
			if (p1)
				free(p1);
			p1 = emalloc(p1_alloc * sizeof(FILE_ENTRY *));
		}
		
		ppanel_file->selected = cnt1;
		for (i = 0; i < cnt1; i++) {
			p1[i] = ppanel_file->files[i];
			p1[i]->select = 1;
		}

		qsort(p1,cnt1,sizeof(FILE_ENTRY *),qcmp);
	}

	ppanel_file->other->selected = ppanel_file->other->pd->cnt;
	for (i = 0; i < ppanel_file->other->pd->cnt; i++) {
		pfe2 = ppanel_file->other->files[i];
		pfe2->select = 1;

		if (panel_cmp_sum.names == cnt1)
			/* we have seen all files from panel#1 */
			continue;	/* not break */

		name2 = SDSTR(pfe2->file);
		for (pfe1 = 0, min = 0, max = cnt1 - 1; min <= max; ) {
			med = (min + max) / 2;
			cmp = strcmp(name2,SDSTR(p1[med]->file));
			if (cmp == 0) {
				pfe1 = p1[med];
				/* entries *pfe1 and *pfe2 have the same name */
				break;
			}
			if (cmp < 0)
				max = med - 1;
			else
				min = med + 1;
		}
		if (pfe1 == 0)
			continue;

		panel_cmp_sum.names++;

		/* all levels: comparing type */
		if (pfe1->file_type == FT_NA || !(
			(IS_FT_PLAIN(pfe1->file_type) && IS_FT_PLAIN(pfe2->file_type))
			|| (IS_FT_DIR(pfe1->file_type) && IS_FT_DIR(pfe2->file_type))
			|| (pfe1->file_type == pfe2->file_type) )
		  )
			continue;

		/* level 1+: comparing size (or device numbers) */
		if (level >= 1
		  && ((IS_FT_DEV(pfe1->file_type) && pfe1->devnum != pfe2->devnum)
		  || (IS_FT_PLAIN(pfe1->file_type) && pfe1->size != pfe2->size)))
			continue;

		/* level 2+: symbolic links not ok */
		if (level >= 2 && pfe1->symlink != pfe2->symlink)
			continue;

		/* level 3,5+: ownership and mode */
		if ((level == 3 || level >= 5) && (pfe1->uid != pfe2->uid
		  || pfe1->gid != pfe2->gid || pfe1->mode12 != pfe2->mode12))
			continue;

		/* level 4+: comparing data (contents) */
		if (level >= 4 && IS_FT_PLAIN(pfe1->file_type)
		  && (cmp = file_cmp(SDSTR(pfe1->file),pathname_join(name2))) ) {
			if (ctrlc_flag)
				break;
			if (cmp < 0 && ++panel_cmp_sum.errors <= 3)
				win_warning_fmt("COMPARE: Cannot read file '%s'.",
				  name2);
			continue;
		}

		/* pair of matching files found */
		pfe1->select = 0;
		ppanel_file->selected--;
		pfe2->select = 0;
		ppanel_file->other->selected--;
		panel_cmp_sum.equal++;
	}

	if (level >= 4)
		signal_ctrlc_off();

	if (ctrlc_flag) {
		win_remark("compare function cancelled");
		/* clear all marks */
		for (i = 0; i < cnt1; i++)
			ppanel_file->files[i]->select = 0;
		ppanel_file->selected = 0;
		for (i = 0; i < ppanel_file->other->pd->cnt; i++)
			ppanel_file->other->files[i]->select = 0;
		ppanel_file->other->selected = 0;

		next_mode = MODE_SPECIAL_RETURN;
		return;
	}

	next_mode = MODE_CMP_SUM;
}

void cx_compare(void)	{ compare_panels(panel_compare.pd->curs); }
